;; Generated package description from org-edna.el
(define-package "org-edna" "1.0alpha1" "Extensible Dependencies 'N' Actions" '((emacs "25.1") (seq "2.19") (org "9.0.5")) :url "https://savannah.nongnu.org/projects/org-edna-el/" :keywords '("convenience" "text" "org"))
